﻿
/*
* Bash transformation for ModSecurity 2.x
* This transformation inhibits several evasion techniques on bash command injection
*
* Copyright (c) 2011-2018 Approach Belgium (https://www.approach.be/)
*
* You may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*     http://www.apache.org/licenses/LICENSE-2.0
*
* If any of the files related to licensing are missing or if you have any
* other questions related to licensing please contact Approach Belgium.
*/

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "apr_fnmatch.h"
#include "apr_optional.h"
#include "util_varbuf.h"
#include "modsecurity.h"
#include <ctype.h>


/*
* delete backslash followed by a newline
* delete backslash escaping next character
* delete single & double quote
* replace all multiple spaces (including tabs) into one space
* Environment variables -> bash behaviour:
*  replace $'xxx' by xxx where ANSI escape sequences are replaced
*  replace $"xxx" by xxx translated according to the current locale
*  replace $xxx by value of xxx
*  replace ${xxx} by value of xxx
*  replace $@, $1, ... $9 by value of params
* as we cannot know the value of an actual environment variable,
* we only delete the complete string as it's used most of the time
* with an inextant variable to produce an empty string.
*/
static int tfn_bash(apr_pool_t *mptmp, char *input,
					long int input_len, char **rval, long int *rval_len)
{
	if (rval == NULL) return -1;
	*rval = (char*)input;
	*rval_len = input_len;
	if (!input_len) return 0;

	int changed = 0, space = 0;
	char *s = input;

	*rval = (char *)input;
	// Check characters
	for (; *input; input++) {
		switch (*input) {
			// remove \ followed by new line
			case '\\':
				if ((input[1] == '\n') || (input[1] == '\r')) {
					input++;
					changed = 1;
					continue;
				}
				if (input[1] == '\\') {
					input++;
					*s++ = '\\';
					changed = 1;
				}
				continue;

				// remove environment variables
			case '$':
			{
				input++;
				changed = 1;

				// remove script parameters
				if (isdigit(*input) || *input == '@') {
					continue;
				}

				char delim = 'a'; // fake delimiter (part of var name)
				switch (*input) {
					case '"':
					case '\'':
						delim = *input++;
						// copy until delimiter
						while (*input != delim) *s++ = *input++;
						continue;
					case '{':
						delim = '}';
						// remove until delimiter
						while (*input != delim) input++;
						continue;
				}

				// remove unquoted var
				while (isalnum(*input) || *input == '_') input++;
				// we went one char too far to test end of var
				input--;
				continue;
			}

			// remove some characters
			case '"':
			case '\'':
				changed = 1;
				continue;

				// replace some characters to space (only one)
			case ' ':
			case '\t':
				if (!space) {
					*s++ = ' ';
					space++;
				}
				changed = 1;
				break;

				// copy normal characters
			default:
				*s++ = *input;
				space = 0;
		}
	}

	*s = 0;
	if (changed) *rval_len = strlen(*rval);
	return changed;
}

static int hook_pre_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_temp)
{
	// Register transformation functions ------------------------------------------
	void(*tfn_fn)(const char *name, void *tfn_fn);
	tfn_fn = APR_RETRIEVE_OPTIONAL_FN(modsec_register_tfn);
	if (!tfn_fn) {
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, NULL, "approach_bash: Unable to find modsec_register_tfn.");
		return DECLINED;
	}

	tfn_fn("bash", (void *)tfn_bash);

	return OK;
}

static void register_hooks(apr_pool_t *p)
{
	ap_hook_pre_config(hook_pre_config, NULL, NULL, APR_HOOK_LAST);
}


/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA approach_bash_module = {
	STANDARD20_MODULE_STUFF,
	NULL,                  /* create per-dir    config structures */
	NULL,                  /* merge  per-dir    config structures */
	NULL,                  /* create per-server config structures */
	NULL,                  /* merge  per-server config structures */
	NULL,                  /* table of config file commands       */
	register_hooks         /* register hooks                      */
};
